# -*- coding: utf-8 -*-
"""tantrum logging module."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import inspect
import logging
import logging.handlers
import time
import six
import sys

from . import exceptions
from . import tools
from .. import __package__ as PACKAGE_ROOT


LOG = logging.getLogger(PACKAGE_ROOT)
""":obj:`logging.Logger`: Package logger."""

LOG.setLevel(logging.DEBUG)

LOG_FILE = "{pkg}.log".format(pkg=PACKAGE_ROOT)
""":obj:`str`: Log filename for :func:`add_file` and :func:`remove_file`."""

LOG_CON_FMT = "%(levelname)-8s [%(name)s:%(funcName)s()] %(message)s"
""":obj:`str`: Logging format for :func:`add_stderr` and :func:`add_stdout`."""

LOG_PATH_FMT = (
    "[%(asctime)s] PID:%(process)s %(levelname)-8s [%(name)s:%(funcName)s()] "
    "%(message)s"
)
""":obj:`str`: Logging format for :func:`add_file`."""

LVL_DICT = logging._levelNames if six.PY2 else logging._nameToLevel
LVL_DICT = {k: v for k, v in LVL_DICT.items() if isinstance(k, six.string_types)}
""":obj:`dict`: Logging level names to ints, python version dependent."""

LVL_STR = ", ".join(list(LVL_DICT) + ["OFF"])
""":obj:`str`: CSV str of logging level names."""


def is_enabled(lvl="debug", obj=LOG):
    """Determine if a logger or handler is enabled for a log level.

    Args:
        lvl (:obj:`str` or :obj:`int`, optional):
            Level to check if obj is enabled for.

            Defaults to: "debug".
        obj (:obj:`logging.Logger` or :obj:`logging.Handler`, optional):
            Object to check if enabled for lvl.

            Defaults to: :data:`LOG`.

    Returns:
        :obj:`bool`

    """
    return obj.isEnabledFor(level_int(lvl=lvl))


def is_disabled(obj=LOG):
    """Determine if a logger or handler is disabled.

    Args:
        obj (:obj:`logging.Logger` or :obj:`logging.Handler`, optional):
            Object to check if disabled.

            Defaults to: :data:`LOG`.

    Notes:
        Will traverse up the logger tree to see if any parents are disabled.

    Returns:
        :obj:`bool`

    """
    parent_disabled = is_disabled(obj=obj.parent) if obj.parent else False
    return any([obj.disabled, parent_disabled])


def use_gmt():
    """Set the logging system to use GMT for time strings."""
    logging.Formatter.converter = time.gmtime


def use_localtime():
    """Set the logging system to use local time for time strings."""
    logging.Formatter.converter = time.localtime


def set_level(lvl="info", obj=LOG):
    """Set a logger or handler to a log level.

    Args:
        lvl (:obj:`str` or :obj:`int`, optional):
            Level to set obj to.

            Defaults to: "info".
        obj (:obj:`logging.Logger` or :obj:`logging.Handler`, optional):
            Object to set lvl on.

            Defaults to: :data:`LOG`.

    Notes:
        If lvl is "OFF" the disabled attr on obj will be set to True.

    Returns:
        :obj:`bool`

    """
    lvl = "info" if lvl is None else lvl
    if isinstance(lvl, six.string_types) and lvl.upper() == "OFF":
        obj.disabled = True
    else:
        obj.disabled = False
        obj.setLevel(level_int(lvl=lvl))


def get_obj_log(obj, lvl="debug"):
    """Get a logger object for an object.

    Args:
        obj (:obj:`object`):
            Object to get a logger for using the objects class module and name.
        lvl (:obj:`str` or :obj:`int`, optional):
            Level to set logger created for obj to initially.

            Defaults to: "debug".

    Returns:
        :obj:`logging.Logger`

    """
    cls = obj if inspect.isclass(obj) else obj.__class__
    name = "{}.{}".format(cls.__module__, cls.__name__)
    return get_log(name=name, lvl=lvl)


def get_log(name, lvl="debug"):
    """Get a logger object.

    Args:
        name (:obj:`str`):
            Path of logger to get/create.
        lvl (:obj:`str` or :obj:`int`, optional):
            Level to set logger created using name initially.

            Defaults to: "debug".

    Returns:
        :obj:`logging.Logger`

    """
    obj = logging.getLogger(name)
    set_level(lvl, obj)
    return obj


def level_name(lvl):
    """Get a logging level in str format.

    Args:
        lvl (:obj:`str` or :obj:`int`):
            Level to get str format of.

    Raises:
        :exc:`tantrum.utils.exceptions.ModuleError`:
            If int is not in values of :data:`LVL_DICT` or if str is not in keys of
            :data:`LVL_DICT`.

    Returns:
        :obj:`str`

    """
    if isinstance(lvl, six.string_types) and lvl.upper() in LVL_DICT:
        return lvl.upper()
    if isinstance(lvl, int) and lvl in LVL_DICT.values():
        return logging.getLevelName(lvl)
    error = "Invalid logging level {lvl!r}, must be one of {lvls}"
    error = error.format(lvl=lvl, lvls=list(LVL_DICT.values()))
    raise exceptions.ModuleError(error)


def level_int(lvl):
    """Get a logging level in int format.

    Args:
        lvl (:obj:`str` or :obj:`int`):
            Level to get int format of.

    Raises:
        :exc:`tantrum.utils.exceptions.ModuleError`:
            If int is not in values of :data:`LVL_DICT` or if str is not in keys of
            :data:`LVL_DICT`.

    Returns:
        :obj:`int`

    """
    if isinstance(lvl, int) and lvl in LVL_DICT.values():
        return lvl
    if isinstance(lvl, six.string_types) and lvl.upper() in LVL_DICT:
        return LVL_DICT[lvl.upper()]
    error = "Invalid logging level {lvl!r}, must be one of {lvls}"
    error = error.format(lvl=lvl, lvls=list(LVL_DICT.keys()))
    raise exceptions.ModuleError(error)


def log_str(obj):
    """Get a str format of a logger object showing level and attached handlers.

    Args:
        obj (:obj:`logging.Logger`):
            Logger object to get str format of.

    Returns:
        :obj:`str`

    """
    handlers = get_output_handlers(obj=obj)
    handlers = [handler_str(obj=h) for h in handlers]
    hstr = ("\n  " + "\n  ".join(handlers)) if handlers else "NO OUTPUT HANDLERS"
    lvl = level_name(lvl=obj.level)
    m = "Logger(name='{o}', level='{lvl}', disabled={d}, handlers: {h})"
    return m.format(o=obj.name, lvl=lvl, d=is_disabled(obj=obj), h=hstr)


def handler_str(obj):
    """Get a str format of a handler object showing level.

    Args:
        obj (:obj:`logging.Handler`):
            Handler object to get str format of.

    Returns:
        :obj:`str`

    """
    cls = obj.__class__.__name__
    lvl = level_name(lvl=obj.level)
    m = "{cls}(name='{name}', level='{lvl}')"
    return m.format(lvl=lvl, name=obj.name, cls=cls)


def add_stdout(lvl="info", fmt=LOG_CON_FMT, obj=LOG):
    """Add a STDOUT handler to a logger object.

    Args:
        lvl (:obj:`str` or :obj:`int`, optional):
            Level to set on handler.

            Defaults to: "info".
        fmt (:obj:`str`, optional):
            Formatting string to use for logging.

            Defaults to: :data:`LOG_CON_FMT`.
        obj (:obj:`logging.Logger`, optional):
            Logger object to add handler to.

            Defaults to: :data:`LOG`.

    Notes:
        Will remove handler from obj if exists before adding.

    Returns:
        :obj:`logging.StreamHandler`

    """
    remove_stdout(obj=obj)
    name = "{name}_stdout".format(name=obj.name)
    log_fmt = logging.Formatter(fmt)
    handler = logging.StreamHandler(stream=sys.stdout)
    handler.setFormatter(log_fmt)
    set_level(lvl=lvl, obj=handler)
    handler.name = name
    obj.addHandler(handler)
    lvl = level_name(lvl=lvl)
    log = log_str(obj=obj)
    m = "Started logging to STDOUT at level {lvl!r} on {log}"
    m = m.format(lvl=lvl, log=log)
    LOG.debug(m)
    return handler


def add_stderr(lvl="info", fmt=LOG_CON_FMT, obj=LOG):
    """Add a STDERR handler to a logger object.

    Args:
        lvl (:obj:`str` or :obj:`int`, optional):
            Level to set on handler.

            Defaults to: "info".
        fmt (:obj:`str`, optional):
            Formatting string to use for logging.

            Defaults to: :data:`LOG_CON_FMT`.
        obj (:obj:`logging.Logger`, optional):
            Logger object to add handler to.

            Defaults to: :data:`LOG`.

    Notes:
        Will remove handler from obj if exists before adding.

    Returns:
        :obj:`logging.StreamHandler`

    """
    remove_stderr(obj=obj)
    name = "{name}_stderr".format(name=obj.name)
    handler = logging.StreamHandler(stream=sys.stderr)
    handler.setFormatter(logging.Formatter(fmt))
    set_level(lvl=lvl, obj=handler)
    handler.name = name
    obj.addHandler(handler)
    lvl = level_name(lvl=lvl)
    log = log_str(obj=obj)
    m = "Started logging to STDERR at level {lvl!r} on {log}"
    m = m.format(lvl=lvl, log=log)
    LOG.debug(m)
    return handler


def add_file(
    lvl="debug",
    path=None,
    path_sub="logs",
    path_file=LOG_FILE,
    max_mb=10,
    max_num=5,
    fmt=LOG_PATH_FMT,
    obj=LOG,
):
    """Add a rotating log file handler to a logger object.

    Args:
        lvl (:obj:`str` or :obj:`int`, optional):
            Level to set on handler.

            Defaults to: "debug".
        path (:obj:`str` or :obj:`pathlib.Path`, optional):
            Storage directory to use.

            If empty, resolve path via :func:`tantrum.utils.tools.get_storage_dir`.

            Defaults to: None.
        path_sub (:obj:`str`, optional):
            Sub directory under path that should contain path_file.

            Defaults to: "logs"
        path_file (:obj:`str`, optional):
            Filename to write logs to under path / path_sub.

            Defaults to: :attr:`LOG_FILE`
        max_mb (:obj:`int`, optional):
            Rotate log file when it reaches this many MB.

            Defaults to: 10.
        max_num (:obj:`int`, optional):
            Only keep up to this number of rotated logs.

            Defaults to: 5.
        fmt (:obj:`str`, optional):
            Formatting string to use for logging.

            Defaults to: :data:`LOG_PATH_FMT`.
        obj (:obj:`logging.Logger`, optional):
            Logger object to add handler to.

            Defaults to: :data:`LOG`.

    Notes:
        Will remove handler from obj if exists before adding.

    Returns:
        :obj:`logging.handlers.RotatingFileHandler`

    """
    remove_file(path=path, path_sub=path_sub, path_file=path_file, obj=obj)

    path = tools.get_storage_dir(path=path, path_sub=path_sub, mkdir=True)
    path = path / path_file

    max_bytes = max_mb * 1024 * 1024
    handler = logging.handlers.RotatingFileHandler(
        filename=format(path), maxBytes=max_bytes, backupCount=max_num
    )
    handler.setFormatter(logging.Formatter(fmt))
    set_level(lvl=lvl, obj=handler)
    handler.name = format(path)
    obj.addHandler(handler)
    lvl = level_name(lvl=lvl)
    log = log_str(obj=obj)
    m = (
        "Started logging to file: '{path}' at level {lvl!r} "
        "rotating {cnt} logs every {mb} MB on {log}"
    )
    m = m.format(path=format(path), lvl=lvl, cnt=max_num, mb=max_mb, log=log)
    LOG.debug(m)
    return handler


def remove_stdout(obj=LOG):
    """Remove a STDOUT handler from a logger object.

    Args:
        obj (:obj:`logging.Logger`, optional):
            Logger object to remove handler from.

            Defaults to: :data:`LOG`.

    Returns:
        :obj:`logging.Handler`

    """
    handler = find_handler_by_name(name="{}_stdout".format(obj.name), obj=obj)
    if handler:
        log = log_str(obj=obj)
        m = "Stopped logging to STDOUT on {log}"
        m = m.format(log=log)
        LOG.debug(m)
        remove_handler(handler=handler, obj=obj)
    return handler


def remove_stderr(obj=LOG):
    """Remove a STDERR handler from a logger object.

    Args:
        obj (:obj:`logging.Logger`, optional):
            Logger object to remove handler from.

            Defaults to: :data:`LOG`.

    Returns:
        :obj:`logging.Handler`

    """
    handler = find_handler_by_name(name="{}_stderr".format(obj.name), obj=obj)
    if handler:
        log = log_str(obj=obj)
        m = "Stopped logging to STDERR on logger {log}"
        m = m.format(log=log)
        LOG.debug(m)
        remove_handler(handler=handler, obj=obj)
    return handler


def remove_file(path=None, path_sub="logs", path_file=LOG_FILE, obj=LOG):
    """Remove a STDERR handler from a logger object.

    Args:
        path (:obj:`str` or :obj:`pathlib.Path`, optional):
            Storage directory to use.

            If empty, resolve path via :func:`tantrum.utils.tools.get_storage_dir`.

            Defaults to: None.
        path_sub (:obj:`str`, optional):
            Sub directory under path that should contain path_file.

            Defaults to: "logs"
        path_file (:obj:`str`, optional):
            Filename to write logs to under path / path_sub.

            Defaults to: :attr:`LOG_FILE`
        obj (:obj:`logging.Logger`, optional):
            Logger object to remove handler from.

            Defaults to: :data:`LOG`.

    Returns:
        :obj:`logging.handlers.RotatingFileHandler`

    """
    path = tools.get_storage_dir(path=path, path_sub=path_sub, mkdir=False)
    path = path / path_file

    handler = find_handler_by_name(name=format(path), obj=obj)
    if handler:
        log = log_str(obj=obj)
        m = "Stopped logging to file: '{path}' on {log}"
        m = m.format(path=format(path), log=log)
        LOG.debug(m)
        try:
            handler.stream.close()
        except Exception:  # nosec # pragma: no cover
            pass
        try:
            handler.close()
        except Exception:  # nosec
            pass
        remove_handler(handler=handler, obj=obj)
    return handler


def add_null(obj=LOG):
    """Add a Null handler to a logger object.

    Args:
        obj (:obj:`logging.Logger`, optional):
            Logger object to add handler to.

            Defaults to: :data:`LOG`.

    Returns:
        :obj:`logging.NullHandler`

    """
    remove_null()
    handler = logging.NullHandler()
    handler.name = "null_handler"
    obj.addHandler(handler)
    return handler


def remove_null(obj=LOG):
    """Remove a Null handler from a logger object.

    Args:
        obj (:obj:`logging.Logger`, optional):
            Logger object to remove handler from.

            Defaults to: :data:`LOG`.

    Returns:
        :obj:`logging.NullHandler`

    """
    handler = find_handler_by_name(name="null_handler", obj=obj)
    remove_handler(obj=obj, handler=handler)
    return handler


def remove_handler(handler, obj=LOG):
    """Remove a handler from a logger.

    Args:
        handler (:obj:`logging.Handler`):
            Handler object to add to logger object.
        obj (:obj:`logging.Logger`, optional):
            Logger object to remove handler from.

            Defaults to: :data:`LOG`.

    Returns:
        :obj:`logging.Handler`

    """
    obj.removeHandler(handler)
    return handler


def find_handler_by_name(name, obj=LOG):
    """Find handler object that is attached to a logger.

    Args:
        name (:obj:`str`):
            Name of handler object to find.
        obj (:obj:`logging.Logger`, optional):
            Logger object to find handler in.

            Defaults to: :data:`LOG`.

    Returns:
        :obj:`logging.Handler` or None

    """
    matches = [x for x in obj.handlers if x.name == name]
    return matches[0] if matches else None


def get_output_handlers(lvl=None, obj=LOG):
    """Find handlers attached to a logger that show output at a given level.

    Args:
        lvl (:obj:`str` or :obj:`int`, optional):
            Level to check if handler outputs at.

            Defaults to: None.
        obj (:obj:`logging.Logger`, optional):
            Logger object to find output handlers in.

            Defaults to: :data:`LOG`.

    Notes:
        If lvl is None logging levels of handlers will not be checked and
        the return will be handlers that are not :obj:`logging.NullHandler`.

    Returns:
        :obj:`list` of :obj:`logging.Handler`

    """
    ret = [h for h in obj.handlers if not isinstance(h, logging.NullHandler)]
    if obj.parent:
        ret += get_output_handlers(obj=obj.parent, lvl=lvl)
    if lvl:
        ret = [x for x in ret if x.level and x.level <= level_int(lvl=lvl)]
    return ret


def will_print_at(lvl="debug", obj=LOG):
    """Check if a logger will show output at a given level.

    Args:
        lvl (:obj:`str` or :obj:`int`, optional):
            Level to check if obj has any handlers that will output logs at.

            Defaults to: "debug".
        obj (:obj:`logging.Logger`, optional):
            Logger to check if disabled and for handlers that output at lvl.

            Defaults to: :data:`LOG`.

    Returns:
        :obj:`bool`

    """
    has_handler = get_output_handlers(obj=obj, lvl=lvl)
    enabled = is_enabled(lvl=lvl, obj=obj)
    disabled = is_disabled(obj=obj)
    if not has_handler or disabled or not enabled:
        return False
    return True


add_null()
# Add a null handler by default to silence logging system warnings
use_gmt()
# Use GMT by default for logging system



# -*- coding: utf-8 -*-
"""tantrum tools module."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import base64
from datetime import datetime, timedelta
import os
import six

from .. import __package__ as PACKAGE_ROOT

if six.PY2:
    import pathlib2 as pathlib  # pragma: no cover
else:
    import pathlib

STORAGE_DIR = "{pkg}_local".format(pkg=PACKAGE_ROOT)
""":obj:`str`: Default sub-directory to use for :attr:`STORAGE_PATH`."""

STORAGE_PATH = format(pathlib.Path("~") / STORAGE_DIR)
""":obj:`str`: Default path to use for :func:`get_storage_dir` if path is None.

Will use OS Environment variable "TANTRUM_STORAGE_PATH" if set.
"""


def get_env(key, default):
    """Get an environment variables value.

    Args:
        key (:obj:`str`):
            Name of OS environment variable value to get.
        default (:obj:`str`):
            Default value to return if key is not set or is empty.

    Returns:
        :obj:`str`

    """
    return os.environ.get(key, default) or default


def get_storage_dir(path=None, path_sub="", mkdir=False):
    """Get the directory for tantrum to use for storing local state.

    Args:
        path (:obj:`str` or :obj:`pathlib.Path`, optional):
            Path to use as base storage directory.

            If None, will use OS environment variable "TANTRUM_STORAGE_PATH" if set.

            If None and "TANTRUM_STORAGE_PATH" is not set,
            will use :data:`STORAGE_PATH`.

            Defaults to: None.
        path_sub (:obj:`str` or :obj:`pathlib.Path`, optional):
            Subdirectory to get under path.

            Defaults to: "".
        mkdir (:obj:`bool`, optional):
            Create path and path / sub with perms 0700.

            Defaults to: False.

    Returns:
        :obj:`pathlib.Path`

    """
    path = path or get_env(key="TANTRUM_STORAGE_DIR", default=STORAGE_PATH)
    path = pathlib.Path(path).expanduser().absolute() / path_sub
    if mkdir:
        path.parent.mkdir(mode=0o700, parents=True, exist_ok=True)
        path.mkdir(mode=0o700, parents=True, exist_ok=True)
    return path


def str_to_intfloat(s):
    """Convert str to int or float.

    Args:
        s (:obj:`str`):
            String to convert to float if it has a period or int if not.

    Returns:
        :obj:`int` or :obj:`float`

    """
    if isinstance(s, six.string_types):
        if "." in s and s.replace(".", "").isdigit():
            return float(s)
        if s.isdigit():
            return int(s)
    return s


def secs_ago(secs):
    """Get a datetime object from N seconds ago.

    Args:
        secs (:obj:`int` or :obj:`float`):
            Number of seconds ago to get datetime object of.

    Returns:
        :obj:`datetime.datetime`

    """
    # secs = 0 if secs is None else secs
    # now = datetime.utcnow()
    # delta = datetime.timedelta(seconds=secs)
    now = datetime.utcnow()
    secs = str_to_intfloat(secs)
    return now if secs is None else now - timedelta(seconds=secs)


def secs_age(dt):
    """Get the number of seconds ago of a datetime object.

    Args:
        dt (:obj:`datetime.datetime`):
            Datetime object to calculate age in seconds of.

    Returns:
        :obj:`int`

    """
    # delta = 0 if dt is None else datetime.utcnow() - dt
    return 0 if not dt else (datetime.utcnow() - dt).seconds


def b64_encode(s):
    """Encode s as base64 using :func:`base64.b64encode`.

    Args:
        s (:obj:`str`):
            Text to base64 encode.

    Returns:
        :obj:`str`

    """
    s = format("" if s is None else s) if not isinstance(s, six.string_types) else s
    s = ensure_binary(s)
    s = base64.b64encode(s)
    return ensure_str(s)


def b64_decode(s):
    """Decode text as base64 using :func:`base64.b64decode`.

    Args:
        s (:obj:`str`):
            Text to base64 decode.

    Returns:
        :obj:`str`

    """
    s = ensure_binary(s)
    s = base64.b64decode(s)
    return ensure_str(s)


def trim_txt(txt, limit=10000):
    """Trim a str if it is over n characters.

    Args:
        txt (:obj:`str`):
            String to trim.
        limit (:obj:`int`, optional):
            Number of characters to trim txt to.

            Defaults to: 10000.

    Returns:
        :obj:`str`

    """
    trim_line = "\n... trimmed over {limit} characters".format(limit=limit)
    txt = txt[:limit] + trim_line if len(txt) > limit else txt
    return txt


def get_dict_path(obj, path):
    """Traverse a dict using a / seperated string.

    Args:
        obj (:obj:`dict`):
            Dictionary to traverse using path.
        path (:obj:`str`):
            Nested dictionary keys seperated by / to traverse in obj.

    Raises:
        :exc:`ValueError`:
            If a part of a path can not be found.

    Returns:
        :obj:`object`

    """
    value = obj
    parts = path.strip("/").split("/")
    for idx, key in enumerate(parts):
        try:
            value = value[key]
        except Exception:
            current_parts = parts[:idx]
            current_path = "/".join(current_parts)
            if isinstance(value, dict):
                valid_keys = list(value.keys())
            else:
                value_type = type(value).__name__
                valid_keys = "NONE: key {k!r} is {t}, not dict."
                valid_keys = valid_keys.format(k=current_parts[-1], t=value_type)
            error = [
                "Unable to find key {k!r} in path {p!r}",
                "Valid keys at {c!r}:",
                "{v!r}",
            ]
            error = "\n  ".join(error)
            error = error.format(k=key, p=path, c=current_path, v=valid_keys)
            raise ValueError(error)
    return value


class Timer(object):
    """Context manager to get a human readable elapsed timespan."""

    def __init__(self):
        """Constructor."""
        self.start = datetime.utcnow()
        """:obj:`datetime.datetime`: Start of timespan."""
        self.end = None
        """:obj:`datetime.datetime`: End of timespan."""

    def __call__(self):
        """Return the current time.

        Returns:
            :obj:`datetime.datetime`

        """
        return datetime.utcnow()

    def __enter__(self):
        """Set the start time.

        Returns:
            :obj:`datetime.datetime`

        """
        self.start = datetime.utcnow()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Set :attr:`end` to now."""
        self.end = self()

    def __str__(self):
        """Get elapsed details.

        Returns:
            :obj:`str`

        """
        m = "delta: {t.elapsed_delta}, int: {t.elapsed_int}, str: {t.elapsed}"
        return m.format(t=self)

    def __repr__(self):
        """Get elapsed details.

        Returns:
            :obj:`str`

        """
        return self.__str__()

    @property
    def elapsed_delta(self):
        """Get elapsed time since :attr:`start`.

        Returns:
            :obj:`datetime.timedelta`

        """
        return (self.end or datetime.utcnow()) - self.start

    @property
    def elapsed_int(self):
        """Get elapsed seconds since :attr:`start`.

        Returns:
            :obj:`int`

        """
        return self.elapsed_delta.seconds

    @property
    def elapsed(self, ms=True, short=True):
        """Get elapsed seconds since :attr:`start` in human readable format.

        ms (:obj:`bool`, optional):
            Represent milliseconds separately instead of as fractional seconds.

            Defaults to: True.
        short (:obj:`bool`, optional):
            Return shortened version of minutes, seconds, millseconds.

            Defaults to: True.

        Returns:
            :obj:`str`

        """
        return self.elapsed_delta


def calc_percent(part, whole):
    """Utility method for getting percentage of part out of whole

    Args:
        part (:obj:`int` or :obj:`float`)
        whole (:obj:`int` or :obj:`float`)

    Returns:
        :obj:`float`

    """
    if 0 in [part, whole]:
        return float(0)
    return 100 * (float(part) / float(whole))


def calc_percent_of(percent, whole):
    """Utility method for getting percentage of whole

    Args:
        percent (:obj:`int` or :obj:`float`)
        whole (:obj:`int` or :obj:`float`)

    Returns:
        :obj:`float`

    """
    return (percent * whole) / 100.0


def def_none(v, d):
    """Use a default value of ``d`` if ``v`` is None."""
    return d if v is None else v


# ripped from six 1.12.0 to allow six 1.11.0 to work
def ensure_str(s, encoding="utf-8", errors="strict"):
    """Coerce *s* to `str`.

    For Python 2:
      - `unicode` -> encoded to `str`
      - `str` -> `str`

    For Python 3:
      - `str` -> `str`
      - `bytes` -> decoded to `str`
    """
    if not isinstance(s, (six.text_type, six.binary_type)):
        raise TypeError("not expecting type '%s'" % type(s))
    if six.PY2 and isinstance(s, six.text_type):
        s = s.encode(encoding, errors)
    elif six.PY3 and isinstance(s, six.binary_type):
        s = s.decode(encoding, errors)
    return s


# ripped from six 1.12.0 to allow six 1.11.0 to work
def ensure_binary(s, encoding="utf-8", errors="strict"):
    """Coerce **s** to six.binary_type.

    For Python 2:
      - `unicode` -> encoded to `str`
      - `str` -> `str`

    For Python 3:
      - `str` -> encoded to `bytes`
      - `bytes` -> `bytes`
    """
    if isinstance(s, six.text_type):
        return s.encode(encoding, errors)
    elif isinstance(s, six.binary_type):
        return s
    else:
        raise TypeError("not expecting type '%s'" % type(s))



# -*- coding: utf-8 -*-
"""Exceptions and warnings for :mod:`tantrum.http_client`."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .. import exceptions


class ModuleError(exceptions.PackageError):
    """Parent of all exceptions for :mod:`tantrum.http_client`."""

    pass


class ModuleWarning(exceptions.PackageWarning):
    """Parent of all warnings for :mod:`tantrum.http_client`."""

    pass
